/*****************************************************************************
 * Copyright (C) 2000 Jim Kent.  This source code may be freely used         *
 * for personal, academic, and non-profit purposes.  Commercial use          *
 * permitted only by explicit agreement with Jim Kent (jim_kent@pacbell.net) *
 *****************************************************************************/
#include "common.h"
#include "hash.h"
#include "linefile.h"
#include "gff.h"

void gffGroupFree(struct gffGroup **pGroup)
/* Free up a gffGroup including lineList. */
{
struct gffGroup *group;
if ((group = *pGroup) != NULL)
    {
    slFreeList(&group->lineList);
    freez(pGroup);
    }
}

void gffGroupFreeList(struct gffGroup **pList)
/* Free up a list of gffGroups. */
{
struct gffGroup *el, *next;
for (el = *pList; el != NULL; el = next)
    {
    next = el->next;
    gffGroupFree(&el);
    }
*pList = NULL;
}


void gffFileFree(struct gffFile **pGff)
/* Free up a gff file. */
{
struct gffFile *gff;
if ((gff = *pGff) != NULL)
    {
    freeMem(gff->fileName);
    freeHash(&gff->seqHash);
    freeHash(&gff->sourceHash);
    freeHash(&gff->featureHash);
    freeHash(&gff->groupHash);
    slFreeList(&gff->lineList);
    slFreeList(&gff->seqList);
    slFreeList(&gff->sourceList);
    slFreeList(&gff->featureList);
    gffGroupFreeList(&gff->groupList);
    freez(pGff);
    }
}

int gffLineCmp(const void *va, const void *vb)
/* Compare two gffLines. */
{
const struct gffLine *a = *((struct gffLine **)va);
const struct gffLine *b = *((struct gffLine **)vb);
int diff;

diff = strcmp(a->seq, b->seq);
if (diff == 0)
    diff = a->start - b->start;
return diff;
}



void gffSyntaxError(struct lineFile *lf)
/* Complain about syntax error in GFF file. */
{
errAbort("Bad line %d of %s:\n%s",
    lf->lineIx, lf->fileName, lf->buf + lf->lineStart);
}

int chopByTab(char *line, char **words, int maxWordCount)
/* Chop up line by tabs. Return number of tab separated fields. */
{
int wordCount = 0;
char c;
while (wordCount < maxWordCount)
    {
    words[wordCount++] = line;
    for (;;)
	{
	if ((c = *line) == 0)
	    goto DONE;
	else if (c == '\t')
	    {
	    *line++ = 0;
	    break;
	    }
	else
	    ++line;
	}
    }
DONE:
   return wordCount;
}

static char *gffTnName(char *seqName, char *groupName)
/* Make name that encorperates seq and group names.... */
{
static char nameBuf[512];
char *s;
if (startsWith("gene-", groupName))
    groupName += 5;
if (startsWith("cc_", groupName))
    groupName += 3;
sprintf(nameBuf, "%s.%s", seqName, groupName);
return nameBuf;
}

void gffFileAdd(struct gffFile *gff, char *fileName, int baseOffset)
/* Create a gffFile structure from a GFF file. */
{
/* Open file and do basic allocations. */
struct lineFile *lf = lineFileOpen(fileName, TRUE);
char *line;
int lineSize;

while (lineFileNext(lf, &line, &lineSize))
    {
    if (line[0] != '#')
	{
	int wordCount;
	char *words[9];
	struct hashEl *hel;
	struct gffLine *gl;
	wordCount = chopByTab(line, words, ArraySize(words));
	if (wordCount < 8)
	    gffSyntaxError(lf);
	AllocVar(gl);

	if ((hel = hashLookup(gff->seqHash, words[0])) == NULL)
	    {
	    struct gffSeqName *el;
	    AllocVar(el);
	    hel = hashAdd(gff->seqHash, words[0], el);
	    el->name = hel->name;
	    slAddHead(&gff->seqList, el);
	    }
	gl->seq = hel->name;

	if ((hel = hashLookup(gff->sourceHash, words[1])) == NULL)
	    {
	    struct gffSource *el;
	    AllocVar(el);
	    hel = hashAdd(gff->sourceHash, words[1], el);
	    el->name = hel->name;
	    slAddHead(&gff->sourceList, el);
	    }
	gl->source = hel->name;

	if ((hel = hashLookup(gff->featureHash, words[2])) == NULL)
	    {
	    struct gffFeature *el;
	    AllocVar(el);
	    hel = hashAdd(gff->featureHash, words[2], el);
	    el->name = hel->name;
	    slAddHead(&gff->featureList, el);
	    }
	gl->feature = hel->name;

	if (!isdigit(words[3][0]) || !isdigit(words[4][0]))
	   gffSyntaxError(lf);	
	gl->start = atoi(words[3])-1 + baseOffset;
	gl->end = atoi(words[4]) + baseOffset;
	gl->score = atof(words[5]);
	gl->strand = words[6][0];
	gl->frame = words[7][0];

	if (wordCount >= 9)
	    {
	    char *tnName = gffTnName(gl->seq, words[8]);
	    if ((hel = hashLookup(gff->groupHash, tnName)) == NULL)
		{
		struct gffGroup *group;
		AllocVar(group);
		hel = hashAdd(gff->groupHash, tnName, group);
		group->name = hel->name;
		group->seq = gl->seq;
		group->source = gl->source;
		slAddHead(&gff->groupList, group);
		}
	    gl->group = hel->name;
	    }

	slAddHead(&gff->lineList, gl);
	}
    }
slReverse(&gff->lineList);
slReverse(&gff->groupList);
lineFileClose(&lf);
}

struct gffFile *gffFileNew(char *fileName)
/* Create a new gffFile structure. */
{
struct gffFile *gff;
AllocVar(gff);
gff->fileName = cloneString(fileName);
gff->seqHash = newHash(6);
gff->sourceHash = newHash(6);
gff->featureHash = newHash(6);
gff->groupHash = newHash(12);
return gff;
}

struct gffFile *gffRead(char *fileName)
/* Create a gffFile structure from a GFF file. */
{
struct gffFile *gff = gffFileNew(fileName);
gffFileAdd(gff, fileName, 0);
return gff;
}

static void getGroupBoundaries(struct gffGroup *group)
/* Fill in start, end, strand of group from lines. */
{
struct gffLine *line;
int start = 0x3fffffff;
int end = -start;
line = group->lineList;
group->strand = line->strand;
for (; line != NULL; line = line->next)
    {
    if (start > line->start)
	start = line->start;
    if (end < line->end)
	end = line->end;
    }
group->start = start;
group->end = end;
}

void gffGroupLines(struct gffFile *gff)
/* Group lines of gff file together, in process mofing
 * gff->lineList to gffGroup->lineList. */
{
struct gffLine *line, *nextLine;
struct hash *groupHash = gff->groupHash;
char *groupName;
struct gffGroup *group;
struct gffLine *ungroupedLines = NULL;

for (line = gff->lineList; line != NULL; line = nextLine)
    {
    nextLine = line->next;
    if ((groupName = line->group) != NULL)
	{
	struct hashEl *hel = hashLookup(groupHash, groupName);
	group = hel->val;
	slAddHead(&group->lineList, line);
	}
    else
	{
	slAddHead(&ungroupedLines, line);
	}
    }

/* Restore ungrouped lines to gff->lineList. */
slReverse(&ungroupedLines);
gff->lineList = ungroupedLines;

/* Restore order of grouped lines and fill in start and end. */
for (group = gff->groupList; group != NULL; group = group->next)
    {
    slSort(&group->lineList, gffLineCmp);
    getGroupBoundaries(group);
    }
}

